/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef DATASETCHAIN_H
#define DATASETCHAIN_H

#include <list>

#include <datatypes.h>
#include <pointer_template.h>

#include <qtreeitem.h>

class QTimer;
class QReadWriteLock;

namespace NSCore {

	class DataSetConsumer;
	class DataSetProducer;
	class DataSetChainMember;

	/**
	 * @brief Baseclass for Producers
	 *
	 * Subclass to produce AcquisitionDataSets and feed them into a chain.
	 *
	 * @todo Allow multiple consumers. Needs new internal way of storage to mark
	 * each dataset for the different consumers.
	 * See branches/epos/multiple_costumers_per_producer
	 *
	 * The cpp-file can select between std::queue and my own simplesafequeue.
	 * Both seem equally fast, mine could be extended to be thread-safe with two
	 * locks for first and last element.
	 */
	class DataSetProducer : public QTreeItem
	{
		Q_OBJECT
		public:
			/**
			 * @brief empty constructor
			 */
			DataSetProducer( QObject* =0 );
			/**
			 * @brief destructor
			 *
			 * Deregisters itself at the consumer if there is one.
			 */
			~DataSetProducer();

			/**
			 * @brief Register a consumer
			 *
			 * This registers a DataSetConsumer as this objects consumer.
			 *
			 * The registered consumer gets notified with the
			 * DataSetProducerEvent.
			 */
			void registerConsumer( DataSetChainMember* );
			/**
			 * @brief Unregister a consumer
			 */
			void unregisterConsumer( DataSetChainMember* );

			/**
			 * @brief Create a XML-representation
			 *
			 * Create ( and return ) a XML-representation from which all the settings
			 * can be restored.
			 */
			virtual QString toXML() const =0;

		signals:

			/**
			 * @brief Speed-o-meter
			 *
			 * The enqueue()-function tries to keep track on how many datasets
			 * where sent to the next chain-member and will emit this signal at
			 * (ir-)regular intervals.
			 */
			void countsPerSecond( double );

		protected:
			/**
			 * @brief Add data to the queue
			 *
			 * This adds a new AcqisitionDataSet to the queue. This doesn't
			 * signal the consumers for new data, you have to do this manually.
			 *
			 * @see signalData();
			 */
			void enqueue( AcqisitionDataSet* );

			/**
			 * @brief The queue is empty
			 *
			 * This function is called when the queue is empty and can be used
			 * to send signals, restart filling the queue, etc.
			 */
			virtual void signalEmpty() {};

		private:
			class DataSetProducerPrivate;
			DataSetProducerPrivate* _d;
	};

	/**
	 * @brief Combines DataSetProducer and DataSetConsumer
	 *
	 * This is a convenience class combining DataSetProducer and
	 * DataSetConsumer. Use it to have chain members that consume DataSets and
	 * produce modified ones.
	 */
	class DataSetChainMember : public DataSetProducer
	{
		Q_OBJECT
		/**
		 * DataSetProducer is a friend to access private functions to signal
		 * new data
		 */
		friend class DataSetProducer;

		public:
			/// Defaultconstructor
			DataSetChainMember( QObject* =0 );
			/// Deconstructor
			~DataSetChainMember();

			/**
			 * @brief Register at a producer
			 *
			 * Registers itself at the specified producer and keeps track of it
			 * to unregister upon deletion
			 */
			void registerAtProducer( DataSetProducer* );

			/**
			 * @brief Unregister a producer
			 *
			 * Unregisters a producer from all lists
			 */
			void unregisterProducer( DataSetProducer* );

			/**
			 * @brief Create a XML-representation
			 *
			 * Create ( and return ) a XML-representation from which all the settings
			 * can be restored.
			 */
			virtual QString toXML() const =0;

			/**
			 * @brief intercept to get events
			 */
			bool event( QEvent* );

		protected:
			/**
			 * @brief Data arrived!
			 *
			 * Subclass this function to handle the data. Return true when the
			 * data should be sent to the next in the chain. Otherwise it gets
			 * deleted.
			 */
			virtual bool dataArrived( NSCore::AcqisitionDataSet* ) =0;

		private:
			std::list <DataSetProducer*> _producers;
	};

};

#endif // DATASETCHAIN_H

