/*
    Copyright 2007-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>
                        Nicki Hinsche <nicvok@freenet.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "dopplerplugins.h"
#include <math.h>

TimeConstantAnalyzer::TimeConstantAnalyzer(  NSCore::QTreeItem* p ) : NSEvaluation::Analyzer(  p ) {
}

void TimeConstantAnalyzer::analyze(  NSCore::AcqisitionDataSet& data ) {
	for ( int i=0; i<1/*data.channels()*/; ++i ) {
		//double max = data.channel( i )->result( "max" )->toDouble();
		int max_at = data.channel( i )->result( "max at" )->toInt();
		int n_samples = data.channel( i )->samples();

		///results by linear regression
		double time_constant = lin_regression( max_at, data.channel( i ), n_samples );

		data.channel( i )->addResult( "time constant", time_constant );
		qDebug() << "time constant =" << time_constant;
	}
}

///lineare Regression _> aus y= a*exp(t) wird log(y)= C + t*x fit nach t, wobei x noch einfach die samples sind zB i=250ps

double TimeConstantAnalyzer::lin_regression( int x_max, NSCore::ChannelDataSet* ch, int n_0 ) const {

	int x_start = x_max + 20;     		  //sample position where regression starts

	int n = n_0 - x_max;                            //data.channels size -> how many samples to fit
	double kn = 1.0/n;                              //easier to calculate later

	double sumX = 0.0;
	double sumY = 0.0;

	for(int i=0; i<n; ++i){                     //or for(int i=x_max; i<data.channels(); ++i)
		sumX+= x_start + i;                     //summed samples
		sumY+= log(ch->sample( x_start + i ));  //summed values of the samples
	}
	double sumXX = 0.0;
	double sumXY = 0.0;

	for(int i=0; i<n; ++i){
		sumXX+=(x_start + i)*(x_start + i);
		sumXY+=log(ch->sample( x_start + i ))*(x_start + i);
	}

	double tau = (kn*sumXY - kn*sumX*kn*sumY)/(kn*sumXX-kn*sumX*kn*sumX);
	//double const = kn*sumY - tau*kn*sumX; 

	return tau;

}

IntegrationAnalyzer::IntegrationAnalyzer( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ) {

	_config->addParameter( "low fraction point", 0.2 );
	_config[ "low fraction point" ]->setMin( 0.0 );
	_config[ "low fraction point" ]->setMax( 1.0 );
	_config->addParameter( "high fraction point", 0.9 );
	_config[ "high fraction point" ]->setMin( 0.0 );
	_config[ "high fraction point" ]->setMax( 1.0 );
	_config->addParameter( "expected mean value", 5.0);
	_config[ "expected mean value" ]->setMin( 0.0 );
	_config[ "expected mean value" ]->setMax( 100.0 );
	_config->addParameter( "expected deviation", 5.0 );
	_config[ "expected mean value" ]->setMin( 0.0 );
	_config[ "expected mean value" ]->setMax( 100.0 );

}

void IntegrationAnalyzer::analyze( NSCore::AcqisitionDataSet& data ) {

	for ( int i=0; i<1/*data.channels()*/; ++i ) {
		double max = data.channel( i )->result( "max" )->toDouble();
		int max_at = data.channel( i )->result( "max at" )->toInt();
		double min = data.channel( i )->result( "min" )->toDouble();
		int min_at = data.channel( i )->result( "min at" )->toInt();

		int n_samples = data.channel(i)->samples();
		double _low_fraction_point = _config->getParameter( "low fraction point" )->toDouble();
		double _high_fraction_point = _config->getParameter( "high fraction point" )->toDouble();
		double _expec_mean = _config->getParameter( "expected mean value" )->toDouble();
		double _expec_deviation = _config->getParameter( "expected deviation" )->toDouble();


		///results
		double area = sumup( max_at, min_at, data.channel( i ), n_samples );
		double area_cf = check_cf( _low_fraction_point, _high_fraction_point, max_at, min_at, data.channel( i ), n_samples );
		double area_cfV2 = check_cfV2( _low_fraction_point, _high_fraction_point, max_at,  min_at, _expec_mean, _expec_deviation, data.channel( i ), n_samples );

		data.channel( i )->addResult( "integral max", area );
		data.channel( i )->addResult( "integral cf", area_cf );
		data.channel( i )->addResult( "integral max cf", area_cfV2);

		qDebug() << " min(x,y), max(x,y): " << min << min_at << max << max_at;
		qDebug() << " Integral pulse after max: " << area;
		qDebug() << " Integral pulse after max with respect to CF: " << area_cfV2;
		qDebug() << " Integral only with respect to CF: " << area_cf;
	}
}


double IntegrationAnalyzer::sumup( int x_max, int x_min, NSCore::ChannelDataSet* ch, int n_0 ) const {
		
	int x_start = x_max;			//sample position of maximum where summing starts
	int n = n_0 - x_max;			//data.channels size -> how many samples to fit

	double sumX = 0.0;
	double sumY = 0.0;

	for(int i=0; i<n; ++i){				//or for(int i=x_max; i<data.channels(); ++i)
		sumX+= x_start + i;			//summed samples
		sumY+= (ch->sample( x_start + i ));	//summed values of the samples
	}

	double flaeche = sumY/1000;

	return flaeche;

}

double IntegrationAnalyzer::check_cf( double lfp, double hfp, int x_max, int x_min, NSCore::ChannelDataSet* ch, int n_0 ) const {
		
	double max_val 	= ch->sample( x_max );	//value of maximum
	double min_val 	= ch->sample( x_min );	//value of minimum


	double scale 	= max_val-min_val;	//range between min and max

	int hcf_at=0;
	int lcf_at=0;
	double hcf=0;
	double lcf=0;

	//at which sample raises the pulse-value above the min CF before the maximum

	for(int i=x_max; i<n_0; i++){
		if ( ch->sample( i ) > (scale*hfp+min_val) ) {
				hcf = ch->sample( i );	//high constant fraction
				hcf_at = i;
			}
	}

	//at which sample is the value below high CF after the maximum

	for(int i=0; i<x_max; i++){
		if ( ch->sample( i ) < (scale*lfp+min_val) ) {
				lcf = ch->sample( i );	//low constant fraction
				lcf_at = i;
			}
	}


	double sumX = 0.0;
	double sumY = 0.0;

	for(int i=lcf_at; i<hcf_at; ++i){	//or for(int i=x_max; i<data.channels(); ++i)
		sumX+= i;			//summed samples
		sumY+= (ch->sample( i ));	//summed values of the samples
	}

	double flaeche_wcf = (sumX*sumY)/1000000000; //normalization to get useful numbers with 10^9

	return flaeche_wcf;

}
double IntegrationAnalyzer::check_cfV2( double lfp, double hfp, int x_max, int x_min, double mean, double deviation, NSCore::ChannelDataSet* ch, int n_0 ) const {
		
	double max_val 	= ch->sample( x_max );	//value of maximum
	double min_val 	= ch->sample( x_min );	//value of minimum

	int x_start = x_max;
	int n = n_0 - x_max;	

	double scale 	= max_val-min_val;	//range between min and max

	int hcf_at=0;
	int lcf_at=0;
	double hcf=0;
	double lcf=0;

	//at which sample raises the pulse-value above the min CF before the maximum

	for(int i=x_max; i<n_0; i++){
		if ( ch->sample( i ) > (scale*hfp+min_val) ) {
				hcf = ch->sample( i );	//high constant fraction
				hcf_at = i;
			}
	}

	//at which sample is the value below high CF after the maximum

	for(int i=0; i<x_max; i++){
		if ( ch->sample( i ) < (scale*lfp+min_val) ) {
				lcf = ch->sample( i );	//low constant fraction
				lcf_at = i;
			}
	}


	double sumX = 0.0;
	double sumY = 0.0;

	for(int i=lcf_at; i<hcf_at; ++i){	//or for(int i=x_max; i<data.channels(); ++i)
		sumX+= i;			//summed samples
		sumY+= (ch->sample( i ));	//summed values of the samples
	}

	double sumXX = 0.0;
	double sumYY = 0.0;

	double flaeche_wcf = (sumX*sumY)/1000000000; //normalization to get useful numbers with 10^9

	//only get well shaped pulses

	if(flaeche_wcf < (mean + deviation) && flaeche_wcf > (mean - deviation)){
		for(int i=0; i<n; ++i){
		sumXX+= x_start + i;
		sumYY+= (ch->sample( x_start + i ));
	}
	}else {
		sumXX=NAN;
		sumYY=NAN;
	}

	double flaeche_wcf_V2 = sumYY/1000;


	return flaeche_wcf_V2;
	
}


/**
 * Code to initialize the plugin-factory on load
 */
extern "C" {
	void start() {
		new DopplerAnalyzers_Client();
	}
}
