/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef EVALUATION_PATH_VIEW_H
#define EVALUATION_PATH_VIEW_H

#include <QtGui/QFrame>
#include <QtGui/QHBoxLayout>
#include <QtGui/QVBoxLayout>
#include <QtCore/QPointer>

#include "evaluationpath.h"

/**
 * @brief GUI just for the PluginItem
 */
class EvaluationPathItem_View : public QFrame {
	Q_OBJECT

	public:
		/**
		 * @brief Constructs a view for a PluginItem.
		 *
		 * This contains just the name and parameters.
		 *
		 * And a Delete-button.
		 */
		EvaluationPathItem_View( NSEvaluation::PluginItem*, QWidget* );
		/// destructor
		~EvaluationPathItem_View();

	signals:
		/**
		 * @brief This plugin wants to get deleted.
		 */
		void deletePlugin( NSEvaluation::PluginItem* );

	private slots:
		/// The button was clicked...
		void deletePlugin();

	private:
		NSEvaluation::PluginItem* _item;
};

/**
 * @brief GUI-container for the PluginItem
 *
 * It is holding the EvaluationPathItem_View and a EvaluationPath_ViewItem per
 * Plugins childs.
 */
class EvaluationPath_ViewItem : public QFrame {
	Q_OBJECT

	public:
		/**
		 * @brief Constructs a container for a PluginItem and its childs
		 */
		EvaluationPath_ViewItem( /*NSEvaluation::PluginItem*/NSCore::QTreeItem*, QWidget* );
		/// destructor
		~EvaluationPath_ViewItem();

		/**
		 * @brief A new plugin was created
		 *
		 * This searches recursivly for a view representing the parent. Then the
		 * parent-view has to create a new view for the new plugin.
		 */
		void newPlugin( NSEvaluation::PluginItem* plugin, NSEvaluation::PluginItem* parent );

	signals:
		/**
		 * @brief A new plugin is wanted
		 *
		 * A new plugin of type is wanted as a child of parent.
		 */
		void newPlugin( QString, NSEvaluation::PluginItem* );
		/**
		 * @brief The view wants the plugin to be deleted.
		 */
		void deletePlugin( NSEvaluation::PluginItem* );

	private slots:
		/// The button for new Plugins was clicked...
		void newPlugin();

		/// Tells this view that its plugin was deleted.
		void pluginDeleted();

	private:
		NSEvaluation::PluginItemPointer _item;
		QPointer<EvaluationPathItem_View> _view;
		QList< QPointer<EvaluationPath_ViewItem> > _childs;
		QVBoxLayout* _layout1;
		QHBoxLayout* _layout2;
};

/**
 * @brief GUI for the base EvaluationPath
 *
 * This class contains a full view for an NSEvaluation::EvaluationPath. It
 * creates child widgets for each plugin and allows you to graphically add and
 * remove widgets.
 */
class EvaluationPath_View : public QFrame {
	Q_OBJECT

	public:
		/**
		 * @brief constructor
		 *
		 * Constructs a new View. The path has to be set via setPath().
		 */
		EvaluationPath_View( QWidget* );
		/// destructor
		~EvaluationPath_View();

		/**
		 * @brief Sets the path this view is responsible for
		 *
		 * Everything regarding the widgets for the plugins is done here.
		 */
		void setPath( NSEvaluation::EvaluationPath* );

	public slots:
		/**
		 * @brief Something has changed.
		 *
		 * As we don't know what has changed, we destroy and rebuild the whole
		 * GUI.
		 */
		void pathChanged();

	private slots:
		/**
		 * @brief A child wants a new plugin.
		 *
		 * This one sends the news to the NSEvaluation::EvaluationPath.
		 */
		void newPlugin( QString, NSEvaluation::PluginItem* );

		/**
		 * @brief A plugin wants to be deleted.
		 *
		 * This one sends the news to the NSEvaluation::EvaluationPath.
		 */
		void deletePlugin( NSEvaluation::PluginItem* );
		/**
		 * @brief A new plugin was created
		 *
		 * Now the GUI has to be updated.
		 *
		 * This has to be here and not inside the corresponding PluginItemView,
		 * because only NSEvaluation::EvaluationPath is a QObject and can send
		 * out signals.
		 */
		void newPlugin( NSEvaluation::PluginItem*, NSEvaluation::PluginItem* );

	private:
		NSEvaluation::EvaluationPath* _path;
		QVBoxLayout* _layout;
		EvaluationPath_ViewItem* _view;
};

#endif // EVALUATION_PATH_VIEW_H

