/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "chartview.h"
#include "chartview.moc"

#include <QtCore/QDebug>
#include <QtCore/QModelIndex>
#include <QtCore/QMetaProperty>
#include <QtGui/QItemSelection>
#include <QtGui/QItemSelectionModel>
#include <QtGui/QScrollBar>
#include <QtGui/QPainter>
#include <QtGui/QMenu>
#include <QtGui/QAction>
#include <QtGui/QContextMenuEvent>

#include <math.h>

ChartView::ChartView( QWidget* p )
	: QAbstractItemView( p )
	, _title( "(no title)" )
	, validItems( 0 )
	, xmin( 1000.0 ), xmax( -1000 )
	, ymin( 1000 ), ymax( -1000 )
	, _dirty( true )
	, _logscale( false )
	, _fwhmline( false )
{
	qDebug( "ChartView::ChartView( QWidget* %p )", p );
	horizontalScrollBar()->setRange( 0,0 );
	verticalScrollBar()->setRange( 0,0 );

	QAction* toggleFWHM = new QAction( "Toggle FWHM-line", this );
	toggleFWHM->setCheckable( true );
	connect( toggleFWHM, SIGNAL( triggered( bool ) ), this, SLOT( setFWHMLine( bool ) ) );
	addAction( toggleFWHM );

	QAction* toggleLog = new QAction( "Toggle logarithmic scale", this );
	toggleLog->setCheckable( true );
	connect( toggleLog, SIGNAL( triggered( bool ) ), this, SLOT( setLogarithmicScale( bool ) ) );
	addAction( toggleLog );
}
ChartView::~ChartView() {
	qDebug( "ChartView::~ChartView()" );
}

QRect ChartView::visualRect( const QModelIndex& ) const {
	//qDebug( "ChartView::visualRect( const QModelIndex& )" );
	return QRect();
}
void ChartView::scrollTo( const QModelIndex&, ScrollHint ) {
	qDebug( "ChartView::scrollTo( const QModelIndex&, ScrollHint )" );
}
QModelIndex ChartView::indexAt( const QPoint& point ) const {
	//qDebug( "ChartView::indexAt( const QPoint& (%i,%i) )", point.x(), point.y() );
	return QModelIndex();
}

void ChartView::dataChanged( const QModelIndex& topLeft, const QModelIndex& bottomright ) {
	//qDebug( "ChartView::dataChanged( const QModelIndex&, const QModelIndex& )" );
	//qDebug() << topLeft << endl << bottomright;
	QAbstractItemView::dataChanged( topLeft, bottomright );

	validItems = 0;
	QModelIndex index;
	double time, count;
	int rowstart = 0;
	int rowend = model()->rowCount( rootIndex() )-1;
	if ( ! _dirty && topLeft.isValid() && bottomright.isValid() ) {
		if ( topLeft.row() == bottomright.row() ) // That probably means a single value was edited,
			_dirty = true;                    // so we better recalculate min and max...
		else {                                    //
			rowstart = topLeft.row();         // Else its just a new value and we check wether
			rowend = bottomright.row();       // this one could be new min/max.
		}

	}
	if ( _dirty ) {
		xmin = 1000.0;
		xmax = -1000;
		ymin = 1000;
		ymax = 0;
	}
	for ( int row=rowstart; row<=rowend; ++row ) {
		// Key
		index = model()->index( row, 0, rootIndex() );
		//qDebug() << index;
		time = model()->data( index ).toDouble();
		if ( time < xmin ) xmin = time;
		if ( time > xmax ) xmax = time;
		// Value
		index = model()->index( row, 1, rootIndex() );
		count = model()->data( index ).toDouble();
		if ( count < ymin ) ymin = count;
		if ( count > ymax ) ymax = count;

		validItems++;
		//qDebug( "values are : (%e|%e)", time, count );
		//qDebug( "xmin = %e, xmax = %e, ymin = %e, ymax = %e", xmin, xmax, ymin, ymax );
	}
	//qDebug( "xmin = %e, xmax = %e, ymin = %e, ymax = %e", xmin, xmax, ymin, ymax );

	_dirty = false;
	viewport()->update();
}
void ChartView::rowsInserted( const QModelIndex&, int, int ) {
	//qDebug( "ChartView::rowsInserted( const QModelIndex&, int, int )" );
}
void ChartView::rowsAboutToBeRemoved( const QModelIndex& index, int a, int b ) {
	//qDebug( "ChartView::rowsAboutToBeRemoved( const QModelIndex& index, int %i, int %i )", a, b );
	//qDebug() << index;
	_dirty = true;
}

bool ChartView::edit( const QModelIndex&, EditTrigger, QEvent* ) {
	qDebug( "ChartView::edit( const QModelIndex&, EditTrigger, QEvent* )" );
	return false;
}
QModelIndex ChartView::moveCursor( QAbstractItemView::CursorAction, Qt::KeyboardModifiers ) {
	qDebug( "ChartView::moveCursor( QAbstractItemView::CursorAction, Qt::KeyboardModifiers )" );
	return QModelIndex();
}

int ChartView::horizontalOffset() const {
	//qDebug( "ChartView::horizontalOffset()" );
	return horizontalScrollBar()->value();
}
int ChartView::verticalOffset() const {
	//qDebug( "ChartView::verticalOffset()" );
	return verticalScrollBar()->value();
}

bool ChartView::isIndexHidden( const QModelIndex& ) const {
	qDebug( "ChartView::isIndexHidden( const QModelIndex& )" );
	return false;
}

void ChartView::setSelection( const QRect&, QItemSelectionModel::SelectionFlags ) {
	//qDebug( "ChartView::setSelection( const QRect&, QItemSelectionModel::SelectionFlags )" );
}

void ChartView::mouseReleaseEvent( QMouseEvent* /*ev*/ ) {
	//qDebug() << "ChartView::mouseReleaseEvent(" << ev << ")";
}
void ChartView::paintEvent( QPaintEvent* ) {
	//qDebug( "ChartView::paintEvent( QPaintEvent* )" );
	//qDebug( "xmin = %e, xmax = %e, ymin = %e, ymax = %e", xmin, xmax, ymin, ymax );

	if ( _dirty )
		dataChanged( QModelIndex(), QModelIndex() );

	if ( validItems > 0 && fabs( xmax-xmin )>0 && fabs( ymax-ymin )>0 ) {
		int border = 10;
		double xstep, ystep;
		int xwidth = viewport()->width() - 2 * border;
		int ywidth = -( viewport()->height() - 2 * border );
		//qDebug( "viewport()->height() = %i, 2*border = %i", viewport()->height(), 2*border );
		xstep = xwidth / double( xmax - xmin );
		ystep = ywidth / double( ymax /*- ymin*/ );
		if ( _logscale ) {
			double lymin = 1;
			if ( ymin > 0.0 ) {
				lymin = log10( ymin );
				//qWarning() << endl << " ymin > 0!" << endl;
			}
			ystep = -fabs( ywidth / double( log10( ymax ) /*- lymin*/ ) );
			//qDebug( "ywidth = %i, log10(ymax)-lymin = %e", ywidth, log10( ymax ) - lymin );
			//qDebug( "lymin = %e, lymax = %e", lymin, log10( ymax ) );
		}
		//qDebug( "xstep = %e, ystep = %e", xstep, ystep );

		QPen penBG( QColor( 100,100,100 ) );
		penBG.setWidth( 2 );
		QPen penData( QColor( 255,0,0 ) );
		penData.setWidth( 1 );

		QPainter painter( viewport() );
		painter.translate( border, border-ywidth );
		painter.save();
		// draw border and scale
		painter.setPen( penBG );
		painter.drawRect( 0, 0, xwidth, ywidth );
		painter.setPen( QPen( QColor( 200,200,200 ) ) );
		// Draw y marks and grid
		if ( !_logscale ) {
			for ( int j=5; j>=0; --j ) {
				if ( ymax > pow( 10, j ) ) {
					double ystep2 = ystep*pow( 10,j );
					for ( int i=1; i*pow( 10,j ) < ymax && fabs( ystep2 )>20; ++i ) {
						double y=i*ystep2;
						if ( _logscale )
							y=log10( i*10 )*ystep;
						painter.drawLine( 0, int( y ), xwidth, int( y ) );
						painter.save();
						painter.setPen( QPen( QColor( 100,100,200 ) ) );
						painter.drawLine( 0, int( y ), 10, int( y ) );
						painter.drawText( 11, int( y ) -1, QString( "%1" ).arg( i*pow( 10,j ) ) );
						painter.restore();
					}
				}
			}
			if ( _fwhmline ) {
				double y = ystep * ymax / 2.0;
				painter.save();
				painter.setPen( QColor( 255, 200, 0 ) );
				painter.drawLine( 0, int( y ), xwidth, int( y ) );
				painter.restore();
			}
		}
		else {
			for ( int j=0; j<=5 && pow( 10, j ) < ymax; ++j ) {
				for ( int i=1; i<11; ++i ) {
					double y = log10( i*pow( 10, j ) )*ystep;
					//qDebug() << "y=" << y << "i=" << i << "j=" << j;
					painter.drawLine( 0, int( y ), xwidth, int( y ) );
					painter.save();
					painter.setPen( QPen( QColor( 100,100,200 ) ) );
					painter.drawLine( 0, int( y ), 10, int( y ) );
					if ( i==1 )
						painter.drawText( 11, int( y )-1, QString( "%1" ).arg( i*pow( 10, j ) ) );
					painter.restore();
				}
			}
			if ( _fwhmline ) {
				double y = ystep * log10( ymax / 2.0 );
				painter.save();
				painter.setPen( QColor( 255, 200, 0 ) );
				painter.drawLine( 0, int( y ), xwidth, int( y ) );
				painter.restore();
			}
		}

		// Draw x marks and grid
		double iinc = 1E-6;
		while ( iinc * xstep < 20 )
			iinc*=10;
		for ( double i=0; i<xmax; i+=iinc ) {
			int x = int( ( i-xmin )*xstep );
			painter.drawLine( x, 0, x, ywidth );
			painter.save();
				painter.setPen( QColor( 100,100,200 ) );
				painter.drawLine( x, 0, x, -10 );
				painter.drawText( x, -11, QString( "%1" ).arg( i ) );
			painter.restore();
		}
		for ( double i=0; i>xmin; i-=iinc ) {
			int x = int( ( i-xmin )*xstep );
			painter.drawLine( x, 0, x, ywidth );
			painter.save();
				painter.setPen( QColor( 100,100,200 ) );
				painter.drawLine( x, 0, x, -10 );
				painter.drawText( x, -11, QString( "%1" ).arg( i ) );
			painter.restore();
		}

		painter.setPen( penBG );
		painter.drawRect( 0, 0, xwidth, ywidth );
		painter.restore();

		painter.save();

		// draw data
		painter.setPen( QColor( 255,200,200 ) );
		QPoint oldpoint; // XXX model has to be sorted for this to work...
		for ( int row=0; row < model()->rowCount( rootIndex() ); ++row ) {
			QModelIndex index = model()->index( row, 0, rootIndex() );
			double x = model()->data( index,Qt::DisplayRole ).toDouble() - xmin;
			index = model()->index( row, 1, rootIndex() );
			double y;
			if ( !_logscale )
				y = model()->data( index,Qt::DisplayRole ).toDouble();
			else
				y = log10( model()->data( index,Qt::DisplayRole ).toDouble() /*- ( ymin-1 )*/ );

			if ( y > - INFINITY && ystep > - INFINITY ) {
				QPoint point( int( x*xstep ), int( y*ystep ) );
				if ( row > 0 )
					painter.drawLine( oldpoint, point );

				oldpoint = point;
			}
		}
		painter.setPen( penData );
		for ( int row=0; row < model()->rowCount( rootIndex() ); ++row ) {
			QModelIndex index = model()->index( row, 0, rootIndex() );
			double x = model()->data( index,Qt::DisplayRole ).toDouble() - xmin;
			index = model()->index( row, 1, rootIndex() );
			double y;
			if ( !_logscale )
				y = model()->data( index,Qt::DisplayRole ).toDouble();
			else
				y = log10( model()->data( index,Qt::DisplayRole ).toDouble() /*- ( ymin-1 )*/ );

			if ( y > - INFINITY && ystep > - INFINITY ) {
				QPoint point( int( x*xstep ), int( y*ystep ) );

				painter.drawLine( point+QPoint( 0,3 ), point-QPoint( 0,3 ) );
				painter.drawLine( point+QPoint( 3,0 ), point-QPoint( 3,0 ) );
			}
		}
		painter.restore();

	}

}
void ChartView::resizeEvent( QResizeEvent* ) {
}
void ChartView::scrollContentsBy( int, int ) {
	qDebug( "ChartView::scrollContentsBy( int, int )" );
}

QRegion ChartView::visualRegionForSelection( const QItemSelection& ) const {
	qDebug( "ChartView::visualRegionForSelection( const QItemSelection& )" );
	return QRegion();
}

void ChartView::contextMenuEvent( QContextMenuEvent* ev ) {
	qDebug() << "ChartView::contextMenuEvent(" << ev << ")";
	QMenu::exec( actions(), ev->globalPos() );
}

