/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "resulttospectrum.h"
#include "resulttospectrum.moc"

#include <QtCore/QDebug>
#include <QtCore/QTimer>
#include <QtCore/QThread>
#include <QtXml/QDomDocument>
#include <QtGui/QIcon>

#include <math.h>
#include <mathexpression.h>
#include "spectrummodel.h"

ResultAction::ResultAction( QString name, ResultsToSpectrum* parent )
	: NSCore::MathExpression( parent )
	, _model( 0 )
{
	qDebug() << "ResultAction::ResultAction(" << name << "," << parent << ")";
	setObjectName( name );
}

ResultAction::~ResultAction() {
	qDebug() << "ResultAction::~ResultAction()";
	if ( _model != 0 )
		delete _model;
}

void ResultAction::doYourJobOnData( NSCore::AcqisitionDataSet& dataset ) {
	//qDebug() << "ResultAction::doYourJobOnData(" << &dataset << ")";
	Abstract1DSpectrumModel* model1d =0;
	double x=NAN;
	double y=NAN;
	if ( _model != 0 )
		model1d = dynamic_cast<Abstract1DSpectrumModel*>( _model );
	for ( int i=0; i<childs(); ++i ) {
		NSCore::MathExpression* exp = dynamic_cast<NSCore::MathExpression*>( child( i ) );
		//qDebug() << "ResultAction::doYourJobOnData(" << &dataset << ") exp =" << exp;
		if ( exp ) {
			double ret = exp->value( dataset );
			if ( !isnan( ret ) && model1d )
				model1d->data( ret );
			if ( !isnan( ret ) ) {
				if ( isnan( x ) )
					x = ret;
				else {
					if ( isnan( y ) )
						y = ret;
				}
			}
		}
	}
	TwoDimSpectrumModel* model2d = dynamic_cast<TwoDimSpectrumModel*>( _model );
	if ( model2d && !isnan( x ) && !isnan( y ) )
		model2d->data( x,y );
}

void ResultAction::guiExecAction( QString action, NSCore::GuiServer_Interface* gui ) {
	ResultsToSpectrum* p = dynamic_cast<ResultsToSpectrum*>( parent() );
	if ( p ) {
		if ( action == "Remove" )
			p->removePath( this, gui );
		if ( action == "Add Exp." )
			p->addExpression( this, gui );
	}
}

void ResultAction::deleteMe( NSCore::MathExpression* exp, NSCore::GuiServer_Interface* gui ) {
	qDebug() << "ResultAction::deleteMe(" << exp << "," << gui << ")";
	ResultsToSpectrum* p = dynamic_cast<ResultsToSpectrum*>( parent() );
	if ( p )
		p->removePath( exp, gui );
}

void ResultAction::addChild( NSCore::MathExpression* exp, NSCore::GuiServer_Interface* gui ) {
	qDebug() << "ResultAction::addChild(" << exp << "," << gui << ")";
	ResultsToSpectrum* p = dynamic_cast<ResultsToSpectrum*>( parent() );
	if ( p )
		p->addExpression( exp, gui );
}

QString ResultAction::toXML() const {
	QStringList paths;
	for ( int i=0; i<childs(); ++i ) {
		NSCore::MathExpression* exp = dynamic_cast<NSCore::MathExpression*>( child( i ) );
		if ( exp )
			paths << exp->toXML();
	}
	QString ret;
	ret += "<spectrum name=\"%1\"%2>";
	if ( !paths.isEmpty() ) {
		foreach( QString path, paths ) {
			ret += "<path>" + path + "</path>";
		}
	}
	ret += _model->toXML();
	ret += "</spectrum>";

	QString dim = "";
	if ( _model->dimensions() > 1 )
		dim = QString( " dimensions=\"%1\"" ).arg( _model->dimensions() );

	return ret.arg( objectName() ).arg( dim );
}


Q_DECLARE_METATYPE( QModelIndex )

ResultsToSpectrum::ResultsToSpectrum( QObject* p )
	: NSCore::DataSetChainMember( p )
	, _timer( new QTimer( this ) )
{
	qRegisterMetaType<QModelIndex>();
	_timer->setInterval( 1000 );
	//_timer->setSingleShot( true );
	connect( _timer, SIGNAL( timeout() ), this, SLOT( checkBuffers() ) );
	_timer->start();
}
ResultsToSpectrum::~ResultsToSpectrum() {
	_timer->stop();
	/*foreach( QString key, _spectra.keys() )
		delete _spectra.take( key );*/
}

void ResultsToSpectrum::checkBuffers() {
	while ( !_bufferedpaths.isEmpty() )
		addPath( _bufferedpaths.takeFirst() );
	while ( !_internalParseXMLBuffer.isEmpty() ) {
		QPair<QDomNode, NSCore::QTreeItem*> item = _internalParseXMLBuffer.takeFirst();
		internalParseXML( item.first, item.second );
	}
	while ( !_actionsToRemove.isEmpty() )
		delete _actionsToRemove.takeFirst();
	while ( !_expressionsToRemove.isEmpty() )
		delete _expressionsToRemove.takeFirst();

	if ( ! _timer->isActive() )
		_timer->start();
}

void ResultsToSpectrum::removeAllPaths() {
	for ( int i=0; i<childs(); ++i ) {
		ResultAction* action = dynamic_cast<ResultAction*>( child( i ) );
		if ( action ) {
			_actionsToRemove.push_back( action );
			emit spectrumDeleted( action->objectName() );
		}
	}
}

void ResultsToSpectrum::removePath( ResultAction* action, NSCore::GuiServer_Interface* gui ) {
	bool yes = gui->messageYesNo( "Really?", "Do you really want to delete this Spectrum with its accompanying paths?" );
	if ( yes ) {
		if ( thread() == QThread::currentThread() )
			delete action;
		else
			_actionsToRemove.push_back( action );

		emit spectrumDeleted( action->objectName() );
		//emit numberOfSpectraChanged();
	}
}
void ResultsToSpectrum::removePath( NSCore::MathExpression* exp, NSCore::GuiServer_Interface* gui ) {
	bool yes = gui->messageYesNo( "Really?", "Do you really want to remove this Expression?" );
	if ( yes ) {
		if ( thread() == QThread::currentThread() )
			delete exp;
		else
			_expressionsToRemove.push_back( exp );
	}
}

void ResultsToSpectrum::addPath( QString xml ) {
	//qDebug() << "ResultsToSpectrum::addPath(" << xml << ")";
	QDomDocument doc( "bla" );
	doc.setContent( xml );
	addPath( doc.documentElement() );
}
void ResultsToSpectrum::addPath( QDomNode xml ) {
	qDebug() << "ResultsToSpectrum::addPath(" << &xml << ")";
	if ( xml.nodeName() == "resultstospectrum" ) {
		qDebug() << " xml is a <resultstospectrum />. Good.";
		qDebug() << " Own thread is" << thread() << "Current thread is" << QThread::currentThread();

		if ( thread() == QThread::currentThread() ) {

			QDomNode elem = xml.firstChildElement( "spectrum" );
			while ( ! elem.isNull() ) {
				//qDebug() << " found a spectrum-element";
				QString name = elem.attributes().namedItem( "name" ).nodeValue();
				QDomNode path = elem.firstChildElement( "path" );
				int dimensions = 1;
				if ( elem.attributes().contains( "dimensions" ) )
					dimensions = elem.attributes().namedItem( "dimensions" ).nodeValue().toInt();

				ResultAction* action = new ResultAction( name, this );

				while ( ! path.isNull() ) {
					internalParseXML( path, action );
					path = path.nextSiblingElement( "path" );
				}

				if ( dimensions == 2 )
					action->_model = new TwoDimSpectrumModel( action );
				else
					action->_model = new ClassicSpectrumModel( action );

				action->_model->setObjectName( "Model" );

				path = elem.firstChildElement( "parameter" );
				if ( !path.isNull() ) {
					NSCore::NamedParameterTreeItem::setValueFromNode( *( action->_model->config() ), path );
				}
				emit numberOfSpectraChanged();

				elem = elem.nextSiblingElement( "spectrum" );
			}
		} else {
			_bufferedpaths.push_back( xml );
			qDebug() << " _bufferedpaths.size() is now" << _bufferedpaths.size();
		}

	}
	qDebug() << "ResultsToSpectrum::addPath()  finished!";
}
void ResultsToSpectrum::addExpression( ResultAction* paction, NSCore::GuiServer_Interface* gui ) {
	internalNewExpression( paction, gui );
}
void ResultsToSpectrum::addExpression( NSCore::MathExpression* pexp, NSCore::GuiServer_Interface* gui ) {
	internalNewExpression( pexp, gui );
}
void ResultsToSpectrum::internalNewExpression( NSCore::QTreeItem* p, NSCore::GuiServer_Interface* gui ) {
	QString type = gui->getItem( "Type of the new Item", "Add math expression", NSCore::MathExpressionFromXMLCreator::the()->availableElements() );
	/*QString type = gui->getItem( "Type of the new Item", "<qt><p>Please select the type of the new item:</p><dl>\
	<dt><b>result</b><dd>Get a result from the data.<br>Allows to set name, channel and<br>generation.\
	<dt><b>sampleparameter</b><dd>Get a parameter from the data.<br>Allows to set name and channel,<br>if the channel is -1, the datasets parameter is returned.\
	<dt><b>smb</b><dd><i>Smaller Minus Bigger</i><br>The results of the <i>calculus</i> get<br>subtracted discriminated by the<br><i>selection</i>\
	<dt><b>add</b><dd>Add two (or more) results together.\
	<dt><b>mul</b><dd>Multiplay two (or more) results together.\
	<dt><b>neg</b><dd>Negate the result.\
	<dt><b>constant</b><dd>A simple constant.\
	<dt><b>precision</b><dd>Reduce the precision of the result<br>to the given precision.\
	</dl></qt>", QStringList() << "result" << "sampleparameter" << "smb" << "add" << "mul" << "neg" << "min" << "max" << "constant" << "precision" );*/
	if ( type.isNull() )
		return;
	QDomDocument doc( "bla" );
	doc.setContent( QString( "<%1 />" ).arg( type ) );
	internalParseXML( doc, p );
}

QString ResultsToSpectrum::toXML() const {
	QString ret;
	for ( int i=0; i<childs(); ++i ) {
		ResultAction* action = dynamic_cast<ResultAction*>( child( i ) );
		if ( action )
			ret += action->toXML();
	}
	if ( ret.isEmpty() )
		return "<resultstospectrum />";
	return "<resultstospectrum>"+ret+"</resultstospectrum>";
}

QStringList ResultsToSpectrum::spectra() const {
	QStringList ret;
	for ( int i=0; i<childs(); ++i ) {
		ResultAction* action = dynamic_cast<ResultAction*>( child( i ) );
		if ( action )
			ret << action->objectName();
	}
	qDebug() << "ResultsToSpectrum::spectra() returning: " << ret.join( "," );
	return ret;
}

AbstractSpectrumModel* ResultsToSpectrum::spectrum( QString name ) {
	qDebug() << "ResultsToSpectrum::spectrum(" << name << ")";
	for ( int i=0; i<childs(); ++i ) {
		ResultAction* action = dynamic_cast<ResultAction*>( child( i ) );
		if ( action && action->objectName() == name )
			return action->_model;
	}
	return 0;
}

bool ResultsToSpectrum::dataArrived( NSCore::AcqisitionDataSet* d ) {
	//qDebug() << "ResultsToSpectrum::dataArrived(" << d << ") [defunct]";
	for ( int i=0; i<childs(); ++i ) {
		ResultAction* action = dynamic_cast<ResultAction*>( child( i ) );
		if ( action ) {
			action->doYourJobOnData( *d );
		}
	}
	return true;
}

void ResultsToSpectrum::guiExecAction( QString action, NSCore::GuiServer_Interface* gui ) {
	qDebug() << "ResultsToSpectrum::guiExecAction(" << action << "," << gui << ")";
	if ( action == "Add Spectrum" ) {
		QString name = gui->getString( "Spectrum name", "Please enter the name of the new spectrum", "" );
		if ( name.isNull() )
			return;

		int dimensions = gui->getInt( "Dimensions", "Please enter the number of (base) dimensions of the spectrum. Currently 1 and 2 are supported.", 1, 1, 2 );
		if ( dimensions > 2 )
			return;

		addPath( QString( "<resultstospectrum><spectrum name=\"%1\" dimensions=\"%2\"/></resultstospectrum>" ).arg( name ).arg( dimensions ) );
	}
}

QVariant ResultsToSpectrum::guiIdentifier( int role ) const {
	switch ( role ) {
		case Qt::DisplayRole:
			return "Results2Spectrum";
			break;
		case Qt::DecorationRole:
			return QIcon( ":/icons/kde/cr48-apps-edu_mathematics.png" );
			break;
		default:
			return QTreeItem::guiIdentifier( role );
	}
}

NSCore::MathExpression* ResultsToSpectrum::internalParseXML( QDomNode node, NSCore::QTreeItem* parent ) {
	qDebug() << "ResultsToSpectrum::InternalParseXML(" << &node << "," << parent << ")";
	qDebug() << " Own thread is" << thread() << "Current thread is" << QThread::currentThread();
	/*QString tmp;
	QTextStream str( &tmp );
	node.save( str, 2 );
	qDebug() << tmp;*/

	if ( node.nodeName() == "path" ) {
		node = node.firstChild();
	}

	if ( thread() != QThread::currentThread() ) {
		qDebug() << "Will buffer the instruction, result of this call will be 0";
		_internalParseXMLBuffer.append( QPair<QDomNode,NSCore::QTreeItem*>( node, parent ) );
		return 0;
	}

	NSCore::MathExpression* returnvalue = NSCore::MathExpressionFromXMLCreator::the()->fromXML( node, parent );

	qDebug() << " will return" << returnvalue;

	return returnvalue;
}

