/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "spectrummodel.h"
#include "spectrummodel.moc"

#include <math.h>
#include <QtCore/QDebug>
#include <QtCore/QFile>
#include <QtCore/QTextStream>

AbstractSpectrumModel::AbstractSpectrumModel( QObject* p )
	: QAbstractItemModel( p )
	, _config( new NSCore::NamedParameterTreeItem( "options", QVariant::Invalid, 0 ) )
	, _changed_index_timer( new QTimer( this ) )
	, _autosavetimer( new QTimer( this ) )
{
	_changed_index_timer->start( 1000 );
	connect( _changed_index_timer, SIGNAL( timeout() ), this, SLOT( changedIndexTimerTimeout() ) );

	connect( _autosavetimer, SIGNAL( timeout() ), this, SLOT( autoSave() ) );

	_config->addParameter( "Autosave", QVariant::Bool );
	_config->addParameter( "Autosave/Filename", "autosave-DATE.csv" );
	_config->getParameter( "Autosave/Filename" )->setToolTip( "Variables to be extended: DATE - current date-time combination, MTIME - start time of the spectrum since last reset or since creation of the spectrum." );
	_config->addParameter( "Autosave/Interval", 10 );
	_config->getParameter( "Autosave/Interval" )->setRange( 1, 60 );
	connect( _config->getParameter( "Autosave" ), SIGNAL( valueChanged() ), this, SLOT( autoSaveChanged() ) );
}

QModelIndex AbstractSpectrumModel::index( int row, int col, const QModelIndex& index ) const {
	if ( index == QModelIndex() )
		return createIndex( row, col );
	return QModelIndex();
}
QModelIndex AbstractSpectrumModel::parent( const QModelIndex& ) const {
	return QModelIndex();
}
QString AbstractSpectrumModel::toXML() const {
	return _config->toXML();
}
void AbstractSpectrumModel::changedIndexTimerTimeout() {
	if ( !_changed_index.isEmpty() ) {
		QMutexLocker lock( &_changed_index_lock );
		qSort( _changed_index );
		emit dataChanged( _changed_index.first(), _changed_index.last() );
		_changed_index.clear();
	}
}

void AbstractSpectrumModel::autoSaveChanged() {
	//qDebug() << "AbstractSpectrumModel::autoSaveChanged()";
	//_config->getParameter( "Autosave" )->debugPrint();
	static bool lastactive = false;
	static int lastinterval = 0;
	if ( lastinterval != _config->getParameter( "Autosave/Interval" )->toInt() ) {
		lastinterval = _config->getParameter( "Autosave/Interval" )->toInt();
		_autosavetimer->stop();
		lastactive = false;
	}
	if ( lastactive != _config->getParameter( "Autosave" )->toBool() ) {
		_autosavetimer->start( lastinterval * 60 * 1000 );
	}
}
void AbstractSpectrumModel::autoSave() {
	//qDebug() << "AbstractSpectrumModel::autoSave()";
	QString filename = _config->getParameter( "Autosave/Filename" )->toString();
	filename.replace( "DATE", QDateTime::currentDateTime().toString( "yyyyMMdd-hhmmss" ) );
	filename.replace( "MTIME", _mtime.toString( "yyyyMMdd-hhmmss" ) );
	saveToFile( filename );
}
Abstract1DSpectrumModel::Abstract1DSpectrumModel( QObject* p ) : AbstractSpectrumModel( p ) {
}
int Abstract1DSpectrumModel::columnCount( const QModelIndex& ) const {
	return 2;
}
QVariant Abstract1DSpectrumModel::headerData( int section, Qt::Orientation orientation, int role ) const {
	if ( role != Qt::DisplayRole )
		return QVariant();

	if ( orientation == Qt::Vertical )
		return section;

	if ( orientation == Qt::Horizontal ) {
		if ( section == 0 )
			return "Value";
		if ( section == 1 )
			return "Count";
	}

	return QVariant();
}


TwoDimSpectrumModel::TwoDimSpectrumModel( QObject* p )
	: AbstractSpectrumModel( p )
	, __rows( 0 )
	, __cols( 0 )
{
	_config->addParameter( "X-Axis", QVariant::Invalid );
	NSCore::NamedParameterTreeItem* xaxis = _config->getParameter( "X-Axis" );
	xaxis->addParameter( "Minimum", -2.0 );
	xaxis->getParameter( "Minimum" )->setRange( -INFINITY, INFINITY );
	xaxis->addParameter( "Maximum", 8.0 );
	xaxis->getParameter( "Maximum" )->setRange( -INFINITY, INFINITY );
	xaxis->addParameter( "Channels", 100 );
	xaxis->getParameter( "Channels" )->setRange( 1, pow( 2,16 ) );
	connect( xaxis, SIGNAL( valueChanged() ), this, SLOT( clearSpectrum() ) );
	_config->addParameter( "Y-Axis", QVariant::Invalid );
	NSCore::NamedParameterTreeItem* yaxis = _config->getParameter( "Y-Axis" );
	yaxis->addParameter( "Minimum", -2.0 );
	yaxis->getParameter( "Minimum" )->setRange( -INFINITY, INFINITY );
	yaxis->addParameter( "Maximum", 8.0 );
	yaxis->getParameter( "Maximum" )->setRange( -INFINITY, INFINITY );
	yaxis->addParameter( "Channels", 100 );
	yaxis->getParameter( "Channels" )->setRange( 1, pow( 2,16 ) );
	connect( yaxis, SIGNAL( valueChanged() ), this, SLOT( clearSpectrum() ) );
	clearSpectrum();
}
int TwoDimSpectrumModel::columnCount( const QModelIndex& ) const {
	return _config->getParameter( "X-Axis/Channels" )->toInt();
}
int TwoDimSpectrumModel::rowCount( const QModelIndex& ) const {
	return _config->getParameter( "Y-Axis/Channels" )->toInt();
}
void TwoDimSpectrumModel::data( double x, double y ) {
	//qDebug() << "TwoDimSpectrumModel::data(" << x << "," << y << ")";
	if ( isnan( x ) || isnan( y ) )
		return;

	double xmin = _config->getParameter( "X-Axis/Minimum" )->toDouble();
	double xmax = _config->getParameter( "X-Axis/Maximum" )->toDouble();
	double ymin = _config->getParameter( "Y-Axis/Minimum" )->toDouble();
	double ymax = _config->getParameter( "Y-Axis/Maximum" )->toDouble();
	if ( !( x < xmax && x > xmin ) )
		return;
	if ( !( y < ymax && y > ymin ) )
		return;

	double xstep = ( xmax - xmin ) / _config->getParameter( "X-Axis/Channels" )->toInt();
	double ystep = ( ymax - ymin ) / _config->getParameter( "Y-Axis/Channels" )->toInt();

	int _x = int( ( x - xmin ) / xstep );
	int _y = int( ( y - ymin ) / ystep );
	_data[ _x ][ _y ] += 1;
	QMutexLocker lock( &_changed_index_lock );
	_changed_index.append( createIndex( _x,_y ) );
}
QVariant TwoDimSpectrumModel::data( const QModelIndex& index, int role ) const {
	if ( role != Qt::DisplayRole )
		return QVariant();

	//qDebug() << "TwoDimSpectrumModel::data(" << index << "," << role << ")";

	if ( index.row() >= _config->getParameter( "Y-Axis/Channels" )->toInt() )
		return QVariant();
	if ( index.column() >= _config->getParameter( "X-Axis/Channels" )->toInt() )
		return QVariant();

	if ( role == Qt::DisplayRole )
		return _data[ index.column() ][ index.row() ];

	return QVariant();
}
QVariant TwoDimSpectrumModel::headerData( int section, Qt::Orientation orientation, int role ) const {
	if ( role != Qt::DisplayRole )
		return QVariant();

	if ( orientation == Qt::Horizontal && section < _config->getParameter( "X-Axis/Channels" )->toInt() )
		return ( _config->getParameter( "X-Axis/Maximum" )->toDouble() - _config->getParameter( "X-Axis/Minimum" )->toDouble() ) / _config->getParameter( "X-Axis/Channels" )->toInt() * section + _config->getParameter( "X-Axis/Minimum" )->toInt();
	if ( orientation == Qt::Vertical && section < _config->getParameter( "Y-Axis/Channels" )->toInt() )
		return ( _config->getParameter( "Y-Axis/Maximum" )->toDouble() - _config->getParameter( "Y-Axis/Minimum" )->toDouble() ) / _config->getParameter( "Y-Axis/Channels" )->toInt() * section + _config->getParameter( "Y-Axis/Minimum" )->toInt();

	return QVariant();
}
void TwoDimSpectrumModel::saveToFile( QString path ) {
	//qDebug() << "TwoDimSpectrumModel::saveToFile( QString" << path << ")";
	double xmin = _config->getParameter( "X-Axis/Minimum" )->toDouble();
	double xmax = _config->getParameter( "X-Axis/Maximum" )->toDouble();
	double ymin = _config->getParameter( "Y-Axis/Minimum" )->toDouble();
	double ymax = _config->getParameter( "Y-Axis/Maximum" )->toDouble();
	double xstep = ( xmax - xmin ) / _config->getParameter( "X-Axis/Channels" )->toInt();
	double ystep = ( ymax - ymin ) / _config->getParameter( "Y-Axis/Channels" )->toInt();
	QFile file( path );
	if ( file.open( QIODevice::WriteOnly ) ) {
		QTextStream stream( &file );

		stream << "ValueX ValueY Count" << endl;
		foreach( int x, _data.keys() ) {
			foreach( int y, _data[ x ].keys() ) {
				stream << xmin + xstep*x << " " << ymin + ystep*y << " " << _data[ x ][ y ] << endl;
			}
		}

		file.close();
	}
}
void TwoDimSpectrumModel::clearSpectrum() {
	//qDebug() << "TwoDimSpectrumModel::clearSpectrum()";
	//beginRemoveColumns( QModelIndex(), 0, __cols );
	beginRemoveRows( QModelIndex(), 0, __rows );
	//_data.clear();
	QList<int> keysx = _data.keys();
	while ( ! _data.isEmpty() ) {
		_data.take( keysx.takeFirst() ).clear();
	}
	//endRemoveColumns();
	endRemoveRows();
	_mtime = QDateTime::currentDateTime();
	__rows = rowCount();
	__cols = columnCount();
	beginInsertRows( QModelIndex(), 0, __rows );
	beginInsertColumns( QModelIndex(), 0, __cols );
	endInsertRows();
	endInsertColumns();
}


SpectrumModel::SpectrumModel( QObject* p ) : Abstract1DSpectrumModel( p ), _accuracy( 0.01 ) {
	_config->addParameter( "accuracy", 0.01 );
	_config->getParameter( "accuracy" )->setMin( 1E-5 );
	_config->getParameter( "accuracy" )->setMax( 1E4 );
	connect( _config, SIGNAL( valueChanged() ), this, SLOT( accuracyChanged() ) );
}
SpectrumModel::~SpectrumModel() {
}

int SpectrumModel::rowCount( const QModelIndex& ) const {
	return _data.size();
}

QVariant SpectrumModel::data( const QModelIndex& index, int role ) const {
	//qDebug() << "SpectrumModel::data( const QModelIndex& " << index << ", int " << role << ")";
	if ( index.isValid() && role == Qt::DisplayRole && index.column() < _data.size() ) {
		if ( index.column() == 0 )
			return _data.at( index.row() ).key;
		else
			return _data.at( index.row() ).count;
	}
	return QVariant();
}

void SpectrumModel::data( double n ) {
	//qDebug() << "SpectrumModel::data( double " << n << " )";
	n = n - fmod( n, _accuracy );
	bool exists = false;
	for ( int i=0; i<_data.size(); ++i ) {
		if ( _data[ i ].key == n ) {
			//qDebug() << "n(" << n << ") exists (index=" << i << ")";
			exists = true;
			_data[ i ].count += 1;
			QModelIndex index0 = createIndex( i, 0 );
			QModelIndex index1 = createIndex( i, 1 );
			emit dataChanged( index0, index1 );
		}
	}
	if ( exists == false ) {
		//qDebug() << "n(" << n << ") doesn't exist creating a new entry";
		int i=0;
		for ( i=0; i<_data.size() && _data[ i ].key < n; ++i ) { }
		beginInsertRows( parent(), i, i );
		_data.insert( i, _Pair( n ) );
		endInsertRows();
		data( n );
	}
}

void SpectrumModel::saveToFile( QString path ) {
	QFile file( path );
	if ( file.open( QIODevice::WriteOnly ) ) {
		QTextStream stream( &file );

		stream << "Value" << " " << "Count" << endl;
		for ( int i=0; i<_data.size(); ++i )
			stream << _data[ i ].key << " " << _data[ i ].count << endl;

		file.close();
	}
}

bool SpectrumModel::removeRows( int row, int count, const QModelIndex& parent ) {
	if ( row+count <= _data.size() ) {
		beginRemoveRows( parent, row, row+count );
		for ( int i=0; i<count; ++i )
			_data.removeAt( row );
		endRemoveRows();
		return true;
	}
	return false;
}

void SpectrumModel::accuracyChanged() {
	_accuracy = config()[ "accuracy" ]->toDouble();
}

ClassicSpectrumModel::ClassicSpectrumModel( QObject* p )
	: Abstract1DSpectrumModel( p )
	, _data_lock( new QReadWriteLock() )
{
	//qDebug() << "ClassicSpectrumModel::ClassicSpectrumModel(" << p << ")";
	_config->addParameter( "Minimum", -2.0 );
	_config->getParameter( "Minimum" )->setMin( -INFINITY );
	_config->getParameter( "Minimum" )->setMax( INFINITY );
	_config->addParameter( "Maximum", 10.0 );
	_config->getParameter( "Maximum" )->setMin( -INFINITY );
	_config->getParameter( "Maximum" )->setMax( INFINITY );
	_config->addParameter( "Channels", 4000 );
	_config->getParameter( "Channels" )->setMin( 1 );
	_config->getParameter( "Channels" )->setMax( pow( 2,16 ) );
	_config->addParameter( "Picoseconds per Sample", 250 );
	_config->getParameter( "Picoseconds per Sample" )->setMin( 125 );
	_config->getParameter( "Picoseconds per Sample" )->setMax( 5000 );
	connect( _config->getParameter( "Minimum" ), SIGNAL( valueChanged() ), this, SLOT( redoChannels() ) );
	connect( _config->getParameter( "Maximum" ), SIGNAL( valueChanged() ), this, SLOT( redoChannels() ) );
	connect( _config->getParameter( "Channels" ), SIGNAL( valueChanged() ), this, SLOT( redoChannels() ) );

	//redoChannels();
	clearSpectrum();
}
ClassicSpectrumModel::~ClassicSpectrumModel() {
	qDebug() << "ClassicSpectrumModel::~ClassicSpectrumModel()";
}

void ClassicSpectrumModel::data( double n ) {
	//qDebug() << "ClassicSpectrumModel::data(" << n << ")";
	int rowc = rowCount();
	for ( int i=0; i<rowc-1; ++i ) {
		QReadLocker lock( _data_lock );
		if ( _data[ i ].key <= n && _data[ i+1 ].key > n ) {
			_data[ i ].count += 1;
			QMutexLocker lock2( &_changed_index_lock );
			_changed_index.push_back( createIndex( i, 1 ) );
			_goodpulses++;
		}
	}
}

QVariant ClassicSpectrumModel::data( const QModelIndex& index, int role ) const {
	//qDebug() << "ClassicSpectrumModel::data( const QModelIndex& " << index << ", int " << role << ")";
	QReadLocker lock( _data_lock );
	if ( index.isValid() && role == Qt::DisplayRole && index.row() < _data.size() ) {
		if ( index.column() == 0 )
			return _data.at( index.row() ).key;
		else
			return _data.at( index.row() ).count;
	}
	return QVariant();
}
void ClassicSpectrumModel::saveToFile( QString path ) {
	qDebug() << "ClassicSpectrumModel::saveToFile(" << path << ")";
	QReadLocker lock( _data_lock );
	QFile file( path );
	if ( file.open( QIODevice::WriteOnly ) ) {
		QTextStream stream( &file );

		if ( path.endsWith( ".dat" ) ) {
			stream << "; Should start with three magic lines and a comment...\r\n";
			// First is class-width
			stream << _config->getParameter( "Picoseconds per Sample" )->toDouble() * fabs( _data[ 0 ].key - _data[ 1 ].key ) / 1000 << "\r\n";
			// Second is temperatur in Kelvin
			stream << 300 << "\r\n";
			// Third is the accuracy of the measurement
			stream << 0.250 << "\r\n";

			for ( int i=0; i<_data.size(); ++i )
				stream << " " << _data[ i ].count << "\r\n";
		}
		else if ( path.endsWith( ".csv" ) ) {
			for ( int i=0; i<_data.size(); ++i )
				stream << _data[ i ].key << " " << _data[ i ].count << endl;
		}

		file.close();
	}
}
void ClassicSpectrumModel::redoChannels() {
	//qDebug() << "Something changed, redoing channels";
	//QMutexLocker lock( &_changed_rows_mutex );
	QWriteLocker lock( _data_lock );

	beginRemoveRows( parent(), 0, _data.size() );
	while ( _data.size() > 0 )
		_data.pop_back();
	endRemoveRows();

	int channels = rowCount();
	double min = config()[ "Minimum" ]->toDouble();
	double max = config()[ "Maximum" ]->toDouble();
	double step = ( max - min ) / channels;

	beginInsertRows( parent(), 0, channels-1 );
	for ( int i=0; i<channels; ++i )
		_data.push_back( _Pair( min + step * i ) );
	endInsertRows();
}

